#!/usr/bin/env node

/**
 * require方法支持的文件类型：.js/.json/.node
 * .js -> module.exports/exports
 * .json -> 通过JSON.parse返回一个对象
 * .node -> process.dlopen (实际开发中基本不用)
 *
 * 特别注意：当接收到的文件类型不是以上三种类型时，比如说是一个file.txt文件，能不能正常执行呢？
 * 答案：以上这种情况，node会把除.js/.json/.node以外的所有文件当作js文件来执行，换句话说只要文件里面的内容符合js语法，就能正常执行！！！！
 */
const exec = require('@foundation-cli/exec');
const path = require('path');
const semver = require('semver'); // 比对版本的模块
const colors = require('colors/safe');
const rootCheck = require('root-check');
const userHome = require('user-home'); // 跨操作系统获取用户主目录
const pathExists = require('path-exists'); // 跨操作系统获取用户主目录
const commander = require('commander');
const dotenv = require('dotenv'); // 加载系统文件里面的环境变量 --- 在系统主目录底下创建一个.env文件，然后这个模块就能读取里面定义的环境变量 --- 具体用法在npm网站里面可以看到
const log = require('@foundation-cli/log');
const {
  getNpmSemverVersion
} = require('@foundation-cli/get-npm-info');
const pkg = require('../package.json');
const constant = require('./constant');

const program = new commander.Command(); // 实例化一个脚手架对象

module.exports = core;

async function core() {
  try {

    await prepare();

    registryCommander();

  } catch (e) {
    log.error(e.message);
  }
}

async function prepare() {
  // 脚手架版本检查
  checkPkgVersion();

  // 检查账户信息
  checkRoot();

  // 获取用户主目录
  checkUserHome();

  // 检查环境变量
  checkEnv();

  // 检查脚手架的版本
  // await checkGlobalUpdate();
}

// 从package.json中获取版本号
function checkPkgVersion() {
  log.info('current_version', pkg.version);
}

// 检查是否使用了root账户启动 （比如说使用sudo执行命令），使用sudo会导致文件权限改变，导致不能删除，不能修改等问题。
// 针对上面说的问题，你就要自动做降级处理
function checkRoot() {
  rootCheck(); // 内部使用process.geteuid()获取当前用户的uid， 如果是root账户（超级管理员），uid就等于0
}

function checkUserHome() {
  if (!userHome || !pathExists.sync(userHome)) {
    throw new Error(colors.red('当前登录用户主目录不存在'));
  }
}

/**
 * 检查环境变量：我们可以把用户名密码这些保存在环境变量中，这样不需要集成在代码中，这样可以实时从环境变量中读取
 */
function checkEnv() {
  const dotenvPath = path.resolve(userHome, './.env');

  if (pathExists.sync(dotenvPath)) {
    const config = dotenv.config({
      path: dotenvPath // 定义.env文件的所在路径，不传的话是当前文件所在的路径
    });

    createDefaultCliConfig();
  }
}

/**
 * 创建默认的环境变量
 */
function createDefaultCliConfig() {
  const cliConfig = {};

  if (process.env.CLI_HOME) {
    cliConfig['cliHome'] = path.join(userHome, process.env.CLI_HOME);
  } else {
    cliConfig['cliHome'] = path.join(userHome, constant.DEFAULT_CLI_HOME);
  }

  process.env.CLI_HOME_PATH = cliConfig.cliHome; // 设置CLI_HOME_PATH环境变量, 这个环境变量主要是来指定本地缓存的地址
}

/**
 * 检查当前版本的脚手架是否需要进行全局更新
 */
async function checkGlobalUpdate() {
  // 1、获取当前版本号和模块名
  const currentVersion = pkg.version;
  const npmName = pkg.name;

  /**
   * 2、调用npm API, 获取所有版本号
   *  API细节：
   *    API地址：https://registry.npmjs.org/${项目名称 --- 比如说 @foundation-cli/core}
   *    method: get
   */
  const lastVersion = await getNpmSemverVersion(currentVersion, npmName);

  // 3、获取最新的版本号，提示用户更新到最新版本号
  if (lastVersion && semver.gt(lastVersion, currentVersion)) {
    log.warn('更新提示：' + colors.yellow(`请手动更新${npmName}，当前版本为：${currentVersion}，最新版本为：${lastVersion}，更新命令为：npm install -g ${npmName}`));
  }
}

/**
 * 注册脚手架
 */
function registryCommander() {
  program
    .name(Object.keys(pkg.bin)[0])
    .usage('<command> [options]')
    .version(pkg.version) // 注册版本
    .option('-d, --debug', '是否开启调试模式', false) // 定义全局的option
    .option('-tp, --targetPath <targetPath>', '是否指定本地调试文件路径', '') // 后面不加<targetPath>的话，默认成为一个Boolean类型的option
    .option('-tp, --targetModule <targetModule>', '是否指定运行模块', ''); // 后面不加<targetPath>的话，默认成为一个Boolean类型的option

  program
    .command('init [projectName]') // projectName是初始化项目的名称
    .option('-f, --force', '是否强制初始化项目') // 如果已经有文件会强制删除，然后再进行初始化
    .action(exec);

  program
    .command('offline-publish')
    .option('-u, --user <user>', '用户名')
    .option('-p, --password <password>', '用户密码')
    .option('-pn, --publishName <publishName>', '发布离线包的用户名称')
    .action(exec);

  /**
   * 注意点：所有的事件监听都是在业务逻辑之前执行，也就是所有命令真正执行之前执行
   */
  // 将checkArgs方法检查debug参数的功能，通过program的事件监听的方式进行转换
  program.on('option:debug', function() {
    if (this.opts().debug) { // program.debug为true，说明用户加入了 --debug 参数
      process.env.LOG_LEVEL = 'verbose';
    } else {
      process.env.LOG_LEVEL = 'info';
    }
    log.level = process.env.LOG_LEVEL;

    log.verbose('debug', '已打开debug模式');
  });

  program.on('option:targetPath', function() {
    let targetPath;

    if (targetPath = this.opts().targetPath) { // program.targetPath为true，说明用户加入了 --targetPath 参数
      process.env.CLI_TARGET_PATH = targetPath;
    }
  });
  program.on('option:targetModule', function() {
    let targetModule;

    if (targetModule = this.opts().targetModule) { // program.targetPath为true，说明用户加入了 --targetPath 参数
      process.env.CLI_TARGET_MODULE = targetModule;
    }
  });

  // 监听所有未知命令
  program.on('command:*', function(obj) {
    const availableCommand = program.commands.map(cmd => cmd.name); // 获取所有可用命令的集合

    console.log(colors.red(`未知命令：${obj[0]}`));

    if (availableCommand && availableCommand.length) {
      console.log(colors.red(`可用命令为：${availableCommand.join(', ')}`));
    }
  });

  program.parse(process.argv); // 解析参数, 参数解析完以后会保存到program.args里面

  if (program.args.length < 1) { // 说明没有输入任何命令，像foundation-cli -d其实也属于没有输入任何功能，所以也打出帮助文档
    program.outputHelp(); // 打印帮助文档
    console.log(); // 加一行空行
  }
}
